from sqlpie.model_engine import ModelEngine
from sqlpie.project import Project
from sqlpie.exceptions import BadInputError, UnknownSourceError
import yaml
import dag

class Sqlpie:
	#render single model
	#Sqlpie(model = 'model_1')
	#render single model with payload
	#Sqlpie(model = 'model_1', vars_payload={'key': 'value'}) 
	#render multiple selected models
	#Sqlpie(models = ['model_1', 'model_2']) 
	#render multiple selected models with payload
	#Sqlpie(models = ['model_1', 'model_2'], {'model_1': {'key': 'value'}})
	#render all but model that are passed in the excludes params
	#Sqlpie(excludes = ['model_3', model_4])
	#render all but model that are passed in the excludes params with payloads
	#Sqlpie(excludes = ['model_3', model_4], vars_payload={'model_1': {'key': value}}) 
	#render all models
	#Sqlpie()
	#render all models with payload
	#Sqlpie(vars_payload={'model_1': {'key':'value'}, 'model_2': { 'key': 'value' }})
  #render all models
  #include in client only selected models
	def __init__(self, **kwargs):
		self.models = {}
		self.table_index = {}
		self.sources_conf = self._set_source_conf()
		self.api_data = {'models': {}, 'selected_models': [], 'dag': {} }
		self.args = kwargs
		conf_keys = list(kwargs.keys())
		conf_keys.sort()
		self.dag = dag.DAG()
		self.dag_edges = []
		if not kwargs:
			self.render_all()
		elif conf_keys == ['vars_payload']:
			self.render_all(vars_payload=kwargs['vars_payload'])
		elif conf_keys == ['models']:
			self.render_multiple(models=kwargs['models'])
		elif conf_keys == ['models', 'vars_payload']:
			self.render_multiple(models=kwargs['models'], vars_payload=kwargs['vars_payload'])
		elif conf_keys == ['excludes']:
			self.exclude_and_render(excludes=kwargs['excludes'])
		elif conf_keys.sort() == ['excludes', 'vars_payload']:
			self.exclude_and_render(excludes=kwargs['excludes'], vars_payload=kwargs['vars_payload'])
		else:
			raise BadInputError
		self.api_data['dag'] = self.build_dag_api_data(self.dag)
		self.build_table_index()
		self.api_data['table_index'] = self.table_index
		self.build_viz_api_data()
	
	def _set_source_conf(self):
		sources_config_file = open("./config/sources.yml", "r")
		sources_conf = yaml.load(sources_config_file, Loader=yaml.FullLoader)
		sources_config_file.close()
		return sources_conf

	def render_all(self, vars_payload={}):
		all_models = Project.models()
		self.render_multiple(all_models, vars_payload)

	def exclude_and_render(self, excludes, vars_payload={}):
		all_models = Project.models()
		models_after_exclusion = [i for i in all_models if not i in excludes]
		self.render_multiple(models=models_after_exclusion, vars_payload=vars_payload)

	def render_multiple(self, models, vars_payload={}):
		for model in Project.models():
			if model in vars_payload.keys():
				rendered_model =  self.render_single(model, vars_payload[model])
			else:
				rendered_model = self.render_single(model)
			if model in models:
				self.api_data['selected_models'].append(model)
				self.append_to_dag(rendered_model)

	def render_single(self, model, vars_payload={}):
		model = ModelEngine(model, vars_payload)
		self.models[model.model] = model
		self.api_data['models'][model.model] = self.build_model_api_data(model)
		return model

	def build_model_api_data(self, model):
		api_data = self.build_dag_api_data(model.dag)
		api_data['staging_model'] = model.staging_model
		api_data['sources'] =  model.model_sources
		api_data['rendered_model'] = model.rendered_model
		return api_data

	def build_dag_api_data(self, dag):
		return 	{
							'ind_nodes': dag.ind_nodes(),
							'all_leaves': dag.all_leaves(),
							'graph_object': self.parse_graph_object(dag.graph),
							'dag_topological_sort': dag.topological_sort(),
							'dag_index': self.build_dag_index(dag)
						}

	def append_to_dag(self, model):
		for edge in model.edges:
			source_table = edge[0]
			destination_table = edge[1]
			self.dag.add_node_if_not_exists(source_table)
			self.dag.add_node_if_not_exists(destination_table)
			if edge not in self.dag_edges:
				self.dag_edges.append(edge)
				self.dag.add_edge(source_table, destination_table)

	def build_dag_index(self, dag):
		obj = {}
		for node in dag.topological_sort():
			obj[node] = {'predecessors': dag.predecessors(node), 'downstreams': dag.downstream(node) }
		return obj

	def parse_graph_object(self, graph):
		dict_graph = dict(graph)
		for key in dict_graph.keys():
			dict_graph[key] = list(dict_graph[key])
		return dict_graph

	def build_viz_api_data(self):
		for model_name, model in self.models.items():
			self.api_data['models'][model_name]['viz_data'] = self.generate_viz_data_from_dag(model.dag)
		self.api_data['dag']['viz_data'] = self.generate_viz_data_from_dag(self.dag)

	def generate_viz_data_from_dag(self, dag):
		data_for_viz = []
		for table in dag.topological_sort():
			downstream = dag.downstream(table)
			table_metadata = self.table_index[table]
			for dep_table in downstream:
				dep_table_metadata = self.table_index[dep_table]
				data_for_viz.append({
														'from': table,
														'to': dep_table,
														'weight': 1,
														'custom_field':{
																						'source_schema': table_metadata['schema'], 
																						'destination_schema': dep_table_metadata['schema']
																						}
													})
		return data_for_viz

	def build_table_index(self):
		for model_name, model in self.models.items():
			for table, table_metadata in model.rendered_model.items():
				self.table_index[table] = self.convert_model_metadata(model, table, table_metadata)
				
		for model_name, model in self.models.items():
			for table, table_metadata in model.model_sources.items():
				if table_metadata['source_type'] in ['model', 'staging_model']:
					if not table in self.table_index.keys():
						raise UnknownSourceError(UnknownSourceError.message(table))
				if table_metadata['source_type'] in ['other_model', 'other_staging_model', 'staging_model']:
					if not 'included_in' in self.table_index[table].keys():
						self.table_index[table]['included_in'] = []
					self.table_index[table]['included_in'].append(model_name)
				elif table_metadata['source_type'] == 'source':
					if not table in self.table_index.keys():
						self.table_index[table] = table_metadata
						self.table_index[table]['included_in'] = []
					self.table_index[table]['included_in'].append(model_name)
		
	def convert_model_metadata(self, model, table_name, table_metadata):
		metadata =  {
									'source_name': model.model,
									'schema': table_metadata['execution_metadata']['destination_schema'],
									'table_name': table_name,
									'update_method': table_metadata['execution_metadata']['update_method']
									
								}
		if metadata['schema'] == model.staging_model:
			metadata['source_type'] = 'staging'
		else:
			metadata['source_type'] = 'model'
		return metadata
	
	def render_from_tables(self, tables):
		self.from_tables_dag_edges = []
		self.from_tables_dag = dag.DAG()
		self.build_specific_tables_dag(tables)
		self.api_data['from_tables_dag'] = self.build_dag_api_data(self.from_tables_dag)
		self.api_data['from_tables_dag']['viz_data'] = self.generate_viz_data_from_dag(self.from_tables_dag)
	
	def build_specific_tables_dag(self, tables):
		i = 0
		while i < len(tables):
			downstreams = self.api_data['dag']['dag_index'][tables[i]]['downstreams']
			for downstream_table in downstreams:
				self.append_to_specific_tables_dag(tables[i], downstream_table)
			self.build_specific_tables_dag(downstreams)
			i+=1

	def append_to_specific_tables_dag(self, source_table, destination_table):
		edge = [source_table, destination_table]
		self.from_tables_dag.add_node_if_not_exists(source_table)
		self.from_tables_dag.add_node_if_not_exists(destination_table)
		if edge not in self.from_tables_dag_edges:
			self.from_tables_dag_edges.append(edge)
			self.from_tables_dag.add_edge(source_table, destination_table)





from jinja2 import Environment

class Compiler:

  def __init__(self, query_path, payload = {}):
    self.file = open(query_path, 'r')
    self.raw_template = self.file.read()
    self.file.close()
    self.payload = payload
    self.jinja = Environment()
    self.query_template = self.jinja.from_string(self.raw_template)
    self.query_string = self.query_template.render(self.payload)


# run pip install -r requirements.txt
from sqlpie.compiler import Compiler
from sqlpie.project import Project
import glob
import sys
import yaml
import dag
from os import listdir
from os.path import isfile, join
from sqlpie.exceptions import UnknownSourceError

class ModelEngine:

  def __init__(self, model, vars_payload={}):
    self.all_models = Project.models()
    self.all_staging_models =  Project.staging_models()
    self.model = model
    self._set_source_conf()
    self._set_models_conf()
    self._vars_payload = vars_payload
    self.model_path = self._get_model_path()
    self.staging_model = self._set_staging_table_name()
    self._model_queries_path = glob.glob(self.model_path)
    self._set_payload()
    self.model_sources = {}
    self._current_query = None
    self.rendered_model = {}
    self.edges = []
    self.dag = dag.DAG()
    self._render_model()
  
  def _get_model_path(self):
    return f"./models/{self.model}/*.sql"
  
  def _load_snippets(self):
    path = './snippets'
    sys.path.append(path)
    snippets = [f for f in listdir(path) if isfile(join(path, f))]
    payload = {}
    for snippet in snippets:
      prefix = snippet.split('.')[0]
      suffix = snippet.split('.')[1]
      if suffix == 'py' and prefix != '__init__':
        modname = prefix
        mod = __import__(modname)
        payload[modname] = mod
    return payload
  
  #triggered when the source block is rendered
  def source(self, source_name, table_name):
    source_table = f"{source_name}.{table_name}"
    if source_name == self.model:
      source_name = self.model
      source_schema = source_name
      source_type = 'model'
    elif source_name ==  self.staging_model:
      source_name = self.model
      source_schema = source_name
      source_type = 'staging_model'
    elif source_name in self.all_models:
      source_schema = source_name
      source_type = 'other_model'
    elif source_name in self.all_staging_models:
      source_schema = source_name
      source_type = 'other_stagings_model'
    else:
      try:
        source_schema = self.sources_conf[source_name]['schema']
        source_type = 'source'
      except KeyError:
        raise UnknownSourceError(UnknownSourceError.message(source_name))
      update_method = None
    destination_table =f"{self._execution_metadata['destination_schema']}.{self._execution_metadata['destination_table']}"
    self.model_sources[source_table] = { 
                                          'source_type': source_type,
                                          'source_name': source_name, 
                                          'schema': source_schema,
                                          'table_name': table_name,
                                          'destination_table': destination_table
                                        }
    self.dag.add_node_if_not_exists(destination_table)
    self.dag.add_node_if_not_exists(source_table)
    edge = [source_table, destination_table]
    if edge not in self.edges:
      self.edges.append(edge)
      self.dag.add_edge( source_table, destination_table)
    if source_name in self.sources_conf.keys():
      return f"{self.sources_conf[source_name]['schema']}.{table_name}"
    else:
      return source_table

  #triggered when the config block is rendered
  def query_execution_config(self, **kargs):
    #input validation 
    self._execution_metadata = kargs
    if 'staging' in self._execution_metadata.keys():
      if self._execution_metadata['staging'] == True:
        self._execution_metadata['destination_schema'] = self.staging_model
    else:
      self._execution_metadata['destination_schema'] = self.model
    return ''

  def _update_current_query(self, query):
    self.current_query = query
  
  def _parse_template_query(self, template):
    config = '\n' + template.split('}}')[0] + "}}"
    query = str('}}').join( template.split('}}')[1:])
    return {'config': config, 'query': query}

  def _render_model(self):
    for path in self._model_queries_path:
      self._update_current_query(path)
      rendered_query =  self._render_query(path)
      table_name = f"{self._execution_metadata['destination_schema']}.{self._execution_metadata['destination_table']}"
      self.rendered_model[table_name] = {}
      self.rendered_model[table_name]['rendered_query'] = rendered_query
      query_template = open(path, 'r')
      self.rendered_model[table_name]['template'] = self._parse_template_query(query_template.read())
      query_template.close()
      self.rendered_model[table_name]['execution_metadata'] = self._execution_metadata

  def _render_query(self, path=None):
    rendered_query = Compiler(path, self.payload).query_string[1:]
    return rendered_query
  
  def _set_source_conf(self):
    sources_config_file = open("./config/sources.yml", "r")
    self.sources_conf = yaml.load(sources_config_file, Loader=yaml.FullLoader)
    sources_config_file.close()

  def _set_staging_table_name(self):
    if self.model_config and 'staging_schema' in self.model_config.keys():
      return self.model_config['staging_schema']
    else:
      return f"{self.model}_staging"
  
  def _set_models_conf(self):
    models_config_file = open(f"./models/{self.model}/model_config.yml", "r")
    self.model_config = yaml.load(models_config_file, Loader=yaml.FullLoader)
    models_config_file.close()
  
  def _set_payload(self):
    self.payload = self._load_snippets()
    self.payload['model'] = self.model
    self.payload['vars'] = self._vars_payload
    self.payload['staging_model'] = self.staging_model
    self.payload['config'] = self.query_execution_config
    self.payload['source'] = self.source

  def print_query(self, destination_table):
    print(self.rendered_model[destination_table])
    return self.rendered_model[destination_table]



from sqlpie.client import Sqlpie
import unittest
from unittest.mock import MagicMock

example_model_client = Sqlpie(models=['example_model'], vars_payload={"example_model": {"test_var": "some_value"}})
example_model = example_model_client.models['example_model']
model_with_custom_staging_schema = Sqlpie(models=['model_with_custom_staging_schema']).models['model_with_custom_staging_schema']
all_models = Sqlpie()
all_models_with_payload = Sqlpie(vars_payload={"example_model": {"test_var": "some_value"}})
excludes = Sqlpie(excludes=['model_with_custom_staging_schema'])

class TestClient(unittest.TestCase):

  def test_dag_validity_for_single_model(self):
    self.assertEqual(example_model.dag.validate()[0], True)

  def test_ind_nodes(self):
  	self.assertEqual(example_model.dag.ind_nodes(),  ['example_source_schema.table_1', 'example_source_schema.table_2', 'example_source_schema.table_3'])
  
  def test_all_leaves(self):
  	self.assertEqual(example_model.dag.all_leaves(),  ['example_model.example_3', 'example_model.example_2'])
  
  def test_model_key_names(self):
  	self.assertEqual(list(example_model.rendered_model.keys()), ['example_model_staging.example_1', 'example_model.example_3', 'example_model.example_2'])

  def test_custom_staging_schema(self):
  	self.assertEqual(model_with_custom_staging_schema.staging_model, 'custom_staging_schema')

  def test_snippets_rendering(self):
  	rendered_query = example_model.rendered_model['example_model_staging.example_1']['rendered_query']
  	query = "\nselect *, date_trunc('month', CONVERT_TIMEZONE ('UTC', 'America/New_York', getdate()))::date\nfrom  example_source_schema_public.table_1 table_1\njoin  example_source_schema_public.table_2 table_2\non table_1.id = table_2.id\njoin  example_source_schema_public.table_3 table_3\non table_2.table_3_id = table_3.id"
  	self.assertEqual(rendered_query, query)

  def test_vars_rendering_single_model(self):
    rendered_query = example_model.rendered_model['example_model.example_2']['rendered_query']
    query = "\nselect *, 'some_value'\nfrom  example_model_staging.example_1"
    self.assertEqual(rendered_query, query)

  def test_vars_rendering_all_models(self):
    rendered_query = all_models_with_payload.models['example_model'].rendered_model['example_model.example_2']['rendered_query']
    query = "\nselect *, 'some_value'\nfrom  example_model_staging.example_1"
    self.assertEqual(rendered_query, query)

  def test_model_names_extraction(self):
    self.assertEqual(example_model.all_models, ['model_with_custom_staging_schema', 'example_model'])

  def test_dag_validity_for_all_models_1(self):
    self.assertEqual(all_models.models['example_model'].dag.validate()[0], True)

  def test_dag_validity_for_all_models_2(self):
    self.assertEqual(all_models.models['model_with_custom_staging_schema'].dag.validate()[0], True)

  def test_excluding(self):
    self.assertEqual(excludes.api_data['selected_models'], ['example_model'])

  def test_model_api_data_keys(self):
    keys = [
            'ind_nodes',
            'all_leaves',
            'graph_object',
            'dag_topological_sort',
            'dag_index',
            'staging_model',
            'sources',
            'rendered_model',
            'viz_data'
            ]
    self.assertEqual( list(all_models.api_data['models']['example_model'].keys()) , keys)

  def test_dag_index_structure(self):
    itterable_object =  {'example_source_schema.table_1': {'predecessors': [], 'downstreams': ['example_model_staging.example_1']}, 'example_source_schema.table_2': {'predecessors': [], 'downstreams': ['example_model_staging.example_1']}, 'example_source_schema.table_3': {'predecessors': [], 'downstreams': ['example_model_staging.example_1']}, 'example_model_staging.example_1': {'predecessors': ['example_source_schema.table_1', 'example_source_schema.table_2', 'example_source_schema.table_3'], 'downstreams': ['example_model.example_3', 'example_model.example_2']}, 'example_model.example_3': {'predecessors': ['example_model_staging.example_1'], 'downstreams': []}, 'example_model.example_2': {'predecessors': ['example_model_staging.example_1'], 'downstreams': []}}
    self.assertEqual(sorted(all_models.api_data['models']['example_model']['dag_index'].items()), sorted(itterable_object.items()))
  
  def test_all_models_edges(self):
    output_edges = sorted(all_models.dag_edges)
    edges = sorted([['example_source_schema.table_3', 'custom_staging_schema.example_1'], ['example_model_staging.example_1', 'custom_staging_schema.example_1'], ['example_model.example_3', 'model_with_custom_staging_schema.example_1'], ['custom_staging_schema.example_1', 'model_with_custom_staging_schema.example_1'], ['example_source_schema.table_1', 'example_model_staging.example_1'], ['example_source_schema.table_2', 'example_model_staging.example_1'], ['example_source_schema.table_3', 'example_model_staging.example_1'], ['example_model_staging.example_1', 'example_model.example_3'], ['example_model_staging.example_1', 'example_model.example_2']])
    self.assertEqual(output_edges, edges)

  def test_single_model_edges(self):
    output_edges = sorted(example_model_client.dag_edges)
    edges = sorted([['example_source_schema.table_1', 'example_model_staging.example_1'], ['example_source_schema.table_2', 'example_model_staging.example_1'], ['example_model_staging.example_1', 'example_model.example_3'], ['example_model_staging.example_1', 'example_model.example_2']])

  def test_api_data_structure(self):
    api_data_keys_output = list(all_models.api_data.keys())
    api_data_keys = ['models', 'selected_models', 'dag', 'table_index']
    self.assertEqual(api_data_keys_output, api_data_keys)

  def test_dag_api_data(self):
    dag_api_data_keys_output = list(all_models.api_data['dag'].keys())
    dag_api_data_keys = ['ind_nodes', 'all_leaves', 'graph_object', 'dag_topological_sort', 'dag_index', 'viz_data']
    self.assertEqual(dag_api_data_keys_output, dag_api_data_keys)
  
  def test_table_index_output(self):
    table_index = {'custom_staging_schema.example_1': {'source_name': 'model_with_custom_staging_schema', 'schema': 'custom_staging_schema', 'table_name': 'custom_staging_schema.example_1', 'update_method': 'append', 'source_type': 'staging', 'included_in': ['model_with_custom_staging_schema']}, 'model_with_custom_staging_schema.example_1': {'source_name': 'model_with_custom_staging_schema', 'schema': 'model_with_custom_staging_schema', 'table_name': 'model_with_custom_staging_schema.example_1', 'update_method': 'append', 'source_type': 'model'}, 'example_model_staging.example_1': {'source_name': 'example_model', 'schema': 'example_model_staging', 'table_name': 'example_model_staging.example_1', 'update_method': 'append', 'source_type': 'staging', 'included_in': ['example_model']}, 'example_model.example_3': {'source_name': 'example_model', 'schema': 'example_model', 'table_name': 'example_model.example_3', 'update_method': 'append', 'source_type': 'model', 'included_in': ['model_with_custom_staging_schema']}, 'example_model.example_2': {'source_name': 'example_model', 'schema': 'example_model', 'table_name': 'example_model.example_2', 'update_method': 'append', 'source_type': 'model'}, 'example_source_schema.table_3': {'source_type': 'source', 'source_name': 'example_source_schema', 'schema': 'example_source_schema_public', 'table_name': 'table_3', 'destination_table': 'custom_staging_schema.example_1', 'included_in': ['model_with_custom_staging_schema', 'example_model']}, 'example_source_schema.table_1': {'source_type': 'source', 'source_name': 'example_source_schema', 'schema': 'example_source_schema_public', 'table_name': 'table_1', 'destination_table': 'example_model_staging.example_1', 'included_in': ['example_model']}, 'example_source_schema.table_2': {'source_type': 'source', 'source_name': 'example_source_schema', 'schema': 'example_source_schema_public', 'table_name': 'table_2', 'destination_table': 'example_model_staging.example_1', 'included_in': ['example_model']}}
    table_index_output = all_models.api_data['table_index']
    self.assertEqual(table_index, table_index_output)

  def test_table_index_included_in(self):
    table_3_included_in_output  = all_models.api_data['table_index']['example_source_schema.table_3']['included_in']
    table_3_included_in = ['model_with_custom_staging_schema', 'example_model']
    self.assertEqual(table_3_included_in, table_3_included_in_output)

  def test_staging_source_type(self):
    staging_source_type = all_models.api_data['table_index']['example_model_staging.example_1']['source_type']
    self.assertEqual(staging_source_type, 'staging')

  def test_staging_model_inclusion(self):
    included_in_output = all_models.api_data['table_index']['example_model_staging.example_1']['included_in']
    included_in = ['example_model']
    self.assertEqual(included_in_output, included_in)
  #add selected models test case

if __name__ == '__main__':
  unittest.main()



class BadInputError(Exception):
	@classmethod
	def message(cls, args):
		return f"The combination of the given args {argss} is invalid \n check our documentation at sqlpie.io/docs/sqlpie-lib"	

class UnknownSourceError(Exception):
	@classmethod
	def message(cls, source):
		return f"The source {source} is used but not declared in the sources configuration or in any one of the models"	


# run pip install -r requirements.txt
import glob
import sys
from os import listdir
import yaml

class Project:
	
	@staticmethod
	def models():
		all_paths = Project._all_model_paths()
		return list(map(lambda x: x.split('/')[-1].split('.')[0] , all_paths))

	@staticmethod
	def staging_models():
		staging_models = []
		models = Project.models()
		for model in models:
			model_config = Project.get_model_config(model)
			if model_config and 'staging_schema' in model_config.keys():
				staging_models.append(model_config['staging_schema'])
			else:
				staging_models.append(f"{model}_staging")
		return staging_models

	@staticmethod
	def model_paths():
		all_paths = Project._all_model_paths()
		models_and_paths = list(map(lambda x: {x.split('/')[-1].split('.')[0]: x} , all_paths))
		models_and_paths_list = {}
		for item in models_and_paths:
			model_name = list(item)[0]
			models_and_paths_list[model_name] = item[model_name]
		return models_and_paths_list

	@staticmethod
	def _all_model_paths():
		return glob.glob('./models/*')
	
	@staticmethod
	def model_config_path(model):
		return f"./models/{model}/model_config.yml"
	
	@staticmethod
	def get_model_config(model):
		config_file = open(Project.model_config_path(model), "r")
		model_conf = yaml.load(config_file, Loader=yaml.FullLoader)
		config_file.close()
		return model_conf


import setuptools
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
  long_description = fh.read()

setup_args = dict(
  name='sqlpie',  
  version='0.7.2.3',
  author="Omri Shuva",
  author_email="omrishuva1@gmail.com",
  description="Data lakes Development Framework",
  long_description=long_description,
  long_description_content_type="text/markdown",
  # url="https://github.com/javatechy/dokr",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent"]
)

install_requires = [
    'py-dag',
    'jinja2',
    'glob3',
    'PyYAML',
]

if __name__ == '__main__':
  setup(**setup_args, install_requires=install_requires)